# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from pytorch_lightning import LightningModule
from torch.nn import ReLU, Sigmoid, Linear, Sequential, Softmax
from torch.nn.functional import binary_cross_entropy_with_logits, cross_entropy
from torch.optim import Adam
from torchmetrics import MetricCollection, Accuracy


class FilterNet(LightningModule):
    def __init__(self):
        super().__init__()
        l1 = Linear(4096, 2000)
        l2 = Linear(2000, 3955)
        act = ReLU(inplace=True)

        self.metric = Accuracy()

        self.body = Sequential(l1, act, l2)
        self.policy_head = Sigmoid()

    def forward(self, x):
        return self.policy_head(self.body(x))

    def _predict(self, x):
        boo = self.forward(x) > 0.1
        return boo.float()

    def training_step(self, batch, batch_idx):
        loss, ba = self._loss(batch, batch_idx)
        self.log('loss', loss)
        self.log('ba', ba, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, ba = self._loss(batch, batch_idx)
        self.log('val_loss', loss)
        self.log('val_ba', ba, prog_bar=True)
        return loss

    def test_step(self, batch, batch_idx):
        loss, ba = self._loss(batch, batch_idx)
        self.log('test_loss', loss)
        self.log('test_ba', ba)
        return loss

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=3e-4)
        return optimizer

    def _loss(self, batch, batch_idx):
        x, y = batch
        body = self.body(x)
        y_pred = self.policy_head(body)
        loss = binary_cross_entropy_with_logits(body, y)
        ba = self.metric(y_pred, y)
        return loss, ba


class SorterNet(LightningModule):
    def __init__(self):
        super().__init__()
        l1 = Linear(4096, 8000)
        l2 = Linear(8000, 3955)

        self.act = ReLU(inplace=True)
        self.head = Softmax()

        metrics = MetricCollection({
            str(n): Accuracy(top_k=n)
            for n
            in [1, 5, 10, 25, 50, 100]
        })

        self.train_metrics = metrics.clone(prefix='train_')
        self.valid_metrics = metrics.clone(prefix='valid_')
        self.test_metrics = metrics.clone(prefix='test_')

        self.body = Sequential(l1, self.act, l2)

    def forward(self, x):
        return self.head(self.body(x))

    def training_step(self, batch, batch_idx):
        loss, m = self._loss_metrics(batch, self.train_metrics)
        self.log('train_loss', loss)
        self.log_dict(m, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, m = self._loss_metrics(batch, self.valid_metrics)
        self.log('valid_loss', loss)
        self.log_dict(m, prog_bar=True)
        return loss

    def test_step(self, batch, batch_idx):
        loss, m = self._loss_metrics(batch, self.test_metrics)
        self.log('test_loss', loss)
        self.log_dict(m, prog_bar=True)
        return loss

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=1e-3)
        return optimizer

    def _loss_metrics(self, batch, metrics: MetricCollection):
        x, y = batch
        y_pred = self.body(x)
        loss = cross_entropy(y_pred, y)
        y_pred = self.head(y_pred)
        m = metrics(y_pred, y)
        return loss, m


__all__ = ['FilterNet', 'SorterNet']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import MoleculeContainer, Reactor, smiles
from collections import deque
from io import TextIOWrapper
from pkg_resources import resource_stream
from typing import Set, FrozenSet
from ..abc import SynthonABC


class RolloutSynthon(SynthonABC):
    __slots__ = ('_depth', '_float')
    __bb__ = None
    __reactors__ = None

    def __new__(cls, molecule, *args, **kwargs):
        if cls.__bb__ is None:
            from .rules import rules
            bb = [smiles(x.strip()) for x in TextIOWrapper(resource_stream(__name__, 'data/building_blocks.smiles'))]
            for b in bb:  # recalculate canonic forms. prevent errors when CGRtools rules set changes.
                b.canonicalize()
            cls.__bb__ = frozenset(str(b) for b in bb)
            cls.__reactors__ = tuple((1., Reactor(x, delete_atoms=True, automorphism_filter=False, one_shot=False))
                                     for x in rules)
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, molecule, /):
        super().__init__(molecule)
        self._float = None

    def __call__(self, finish=10, **kwargs):
        self._depth = finish

    def __float__(self):
        if self._float is not None:
            return self._float
        elif self:
            self._float = 1.
            return self._float
        molecule = self._molecule
        seen = set()
        max_depth = self._depth
        queue = deque([(molecule, 0)])
        while queue:
            curr, depth = queue.popleft()
            depth = depth + 1
            if depth > max_depth:
                self._float = -.5
                return self._float
            seen.add(curr)
            try:
                result = next(x for _, r in self._sorted() for x in r([curr])).products
            except StopIteration:
                self._float = -1.
                return self._float
            for mol in result:
                mol.kekule()
                mol.thiele()
            if seen.isdisjoint(result):
                queue.extend((x, depth) for x in result if str(x) not in self.__bb__)
        self._float = 1.
        return self._float

    def __iter__(self):
        if self:
            return
        molecule = self.molecule
        seen: Set[FrozenSet['MoleculeContainer']] = set()
        for prob, reactor in self._sorted():
            for reaction in reactor([molecule]):
                prods = [m for x in reaction.products for m in x.split()]
                for mol in prods:
                    mol.kekule()
                    mol.thiele()
                products = frozenset(mol for mol in prods)
                if products in seen:
                    continue
                seen.add(products)
                yield prob, tuple(type(self)(mol) for mol in products)

    def __bool__(self):
        return str(self._molecule) in self.__bb__

    def _sorted(self):
        return self.__reactors__


__all__ = ['RolloutSynthon']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from CGRtools import MoleculeContainer
from typing import Tuple, Iterator


class SynthonABC(ABC):
    __slots__ = ('_molecule', '_mapping')
    __cache__ = {}

    def __init__(self, molecule: MoleculeContainer, /):
        if str(molecule) in self.__cache__:
            self._molecule = self.__cache__[str(molecule)]
            self._mapping = self._molecule.get_fast_mapping(molecule)
        else:
            self.__cache__[str(molecule)] = molecule
            self._molecule = molecule
            self._mapping = None

    def __call__(self, **kwargs):
        """
        Additional init params for proper synthon working.

        Called by Scroll.
        """

    @abstractmethod
    def __iter__(self) -> Iterator[Tuple[float, Tuple['SynthonABC', ...]]]:
        """
        Generator of precursors synthons. Ordered by preference of reactions.
        The number by which the rules are sorted, has a mathematical meaning of probability to take this reaction.

        :return: Pairs of reaction value and tuple of synthons.
        """

    @abstractmethod
    def __bool__(self):
        """
        Is building block.
        """

    @abstractmethod
    def __float__(self):
        """
        Value of synthesisability.
        Can return a number from -1 to 1.
        Represent an opportunity to synthesize a molecule from commercially available compounds.
        """

    def __len__(self):
        return len(self._molecule)

    def __hash__(self):
        return hash(self._molecule)

    def __eq__(self, other: 'SynthonABC'):
        return self._molecule == other._molecule

    @property
    def molecule(self) -> MoleculeContainer:
        if self._mapping:
            return self._molecule.remap(self._mapping, copy=True)
        return self._molecule.copy()

    def __repr__(self):
        return str(self._molecule)


__all__ = ['SynthonABC']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import QueryContainer, ReactionContainer
from CGRtools.periodictable import ListElement


rules = []


def prepare():
    q_ = QueryContainer()
    p1_ = QueryContainer()
    p2_ = QueryContainer()
    rules.append(ReactionContainer((q_,), (p1_, p2_)))
    return q_, p1_, p2_


# R-amide/ester formation
# [C](-[N,O;D23;Zs])(-[C])=[O]>>[A].[C]-[C](-[O])=[O]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('O')
q.add_atom(ListElement(['N', 'O']), hybridization=1, neighbors=(2, 3))
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 2)
q.add_bond(2, 4, 1)

p1.add_atom('C')
p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('O', _map=5)
p1.add_bond(1, 2, 1)
p1.add_bond(2, 3, 2)
p1.add_bond(2, 5, 1)

p2.add_atom('A', _map=4)


# acyl group addition with aromatic carbon's case (Friedel-Crafts)
# [C;Za]-[C](-[C])=[O]>>[C].[C]-[C](-[Cl])=[O]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('O')
q.add_atom('C', hybridization=4)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 2)
q.add_bond(2, 4, 1)

p1.add_atom('C')
p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('Cl', _map=5)
p1.add_bond(1, 2, 1)
p1.add_bond(2, 3, 2)
p1.add_bond(2, 5, 1)

p2.add_atom('C', _map=4)


# Williamson reaction
# [C;Za]-[O]-[C;Zs;W0]>>[C]-[Br].[C]-[O]
q, p1, p2 = prepare()
q.add_atom('C', hybridization=4)
q.add_atom('O')
q.add_atom('C', hybridization=1, heteroatoms=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)

p1.add_atom('C')
p1.add_atom('O')
p1.add_bond(1, 2, 1)

p2.add_atom('C', _map=3)
p2.add_atom('Br')
p2.add_bond(3, 4, 1)


# Buchwald-Hartwig amination
# [N;D23;Zs;W0]-[C;Za]>>[C]-[Br].[N]
q, p1, p2 = prepare()
q.add_atom('N', heteroatoms=0, hybridization=1, neighbors=(2, 3))
q.add_atom('C', hybridization=4)
q.add_bond(1, 2, 1)

p1.add_atom('C', _map=2)
p1.add_atom('Br')
p1.add_bond(2, 3, 1)

p2.add_atom('N')


# imidazole imine atom's alkylation
# [C;r5](:[N;r5]-[C;Zs;W1]):[N;D2;r5]>>[C]-[Br].[N]:[C]:[N]
q, p1, p2 = prepare()
q.add_atom('N', rings_sizes=5)
q.add_atom('C', rings_sizes=5)
q.add_atom('N', rings_sizes=5, neighbors=2)
q.add_atom('C', hybridization=1, heteroatoms=(1, 2))
q.add_bond(1, 2, 4)
q.add_bond(2, 3, 4)
q.add_bond(1, 4, 1)

p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('N')
p1.add_bond(1, 2, 4)
p1.add_bond(2, 3, 4)

p2.add_atom('C', _map=4)
p2.add_atom('Br')
p2.add_bond(4, 5, 1)


# Knoevenagel condensation (nitryl and carboxyl case)
# [C]=[C](-[C]#[N])-[C](-[O])=[O]>>[C]=[O].[C](-[C]#[N])-[C](-[O])=[O]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('N')
q.add_atom('C')
q.add_atom('O')
q.add_atom('O')
q.add_bond(1, 2, 2)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 3)
q.add_bond(2, 5, 1)
q.add_bond(5, 6, 2)
q.add_bond(5, 7, 1)

p1.add_atom('C', _map=2)
p1.add_atom('C')
p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('O')
p1.add_bond(2, 3, 1)
p1.add_bond(3, 4, 3)
p1.add_bond(2, 5, 1)
p1.add_bond(5, 6, 2)
p1.add_bond(5, 7, 1)

p2.add_atom('C', _map=1)
p2.add_atom('O', _map=8)
p2.add_bond(1, 8, 2)


# Knoevenagel condensation (double nitryl case)
# [C]=[C](-[C]#[N])-[C]#[N]>>[C]=[O].[C](-[C]#[N])-[C]#[N]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('N')
q.add_atom('C')
q.add_atom('N')
q.add_bond(1, 2, 2)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 3)
q.add_bond(2, 5, 1)
q.add_bond(5, 6, 3)

p1.add_atom('C', _map=2)
p1.add_atom('C')
p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('N')
p1.add_bond(2, 3, 1)
p1.add_bond(3, 4, 3)
p1.add_bond(2, 5, 1)
p1.add_bond(5, 6, 3)

p2.add_atom('C', _map=1)
p2.add_atom('O', _map=8)
p2.add_bond(1, 8, 2)


# Knoevenagel condensation (double carboxyl case)
# [C]=[C](-[C](-[O])=[O])-[C](-[O])=[O]>>[C]=[O].[C](-[C](-[O])=[O])-[C](-[O])=[O]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('O')
q.add_atom('O')
q.add_atom('C')
q.add_atom('O')
q.add_atom('O')
q.add_bond(1, 2, 2)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 2)
q.add_bond(3, 5, 1)
q.add_bond(2, 6, 1)
q.add_bond(6, 7, 2)
q.add_bond(6, 8, 1)

p1.add_atom('C', _map=2)
p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('O')
p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('O')
p1.add_bond(2, 3, 1)
p1.add_bond(3, 4, 2)
p1.add_bond(3, 5, 1)
p1.add_bond(2, 6, 1)
p1.add_bond(6, 7, 2)
p1.add_bond(6, 8, 1)

p2.add_atom('C', _map=1)
p2.add_atom('O', _map=9)
p2.add_bond(1, 9, 2)


# heterocyclization with guanidine
# [c]((-[N;W0;Zs])@[n]@[c](-[N;D1])@[c;W0])@[n]@[c]-[O; D1]>>[C](-[N])(=[N])-[N].[C](#[N])-[C]-[C](-[O])=[O]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('N', heteroatoms=0, hybridization=1)
q.add_atom('N')
q.add_atom('C')
q.add_atom('N', neighbors=1)
q.add_atom('C', heteroatoms=0)
q.add_atom('N')
q.add_atom('C')
q.add_atom('O', neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(1, 3, 4)
q.add_bond(3, 4, 4)
q.add_bond(4, 5, 1)
q.add_bond(4, 6, 4)
q.add_bond(1, 7, 4)
q.add_bond(7, 8, 4)
q.add_bond(8, 9, 1)

p1.add_atom('C')
p1.add_atom('N')
p1.add_atom('N')
p1.add_atom('N', _map=7)
p1.add_bond(1, 2, 1)
p1.add_bond(1, 3, 2)
p1.add_bond(1, 7, 1)

p2.add_atom('C', _map=4)
p2.add_atom('N')
p2.add_atom('C')
p2.add_atom('C', _map=8)
p2.add_atom('O', _map=9)
p2.add_atom('O')
p2.add_bond(4, 5, 3)
p2.add_bond(4, 6, 1)
p2.add_bond(6, 8, 1)
p2.add_bond(8, 9, 2)
p2.add_bond(8, 10, 1)


# alkylation of amine
# [C]-[N]-[C]>>[C]-[N].[C]-[Br]
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('N')
q.add_atom('C')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(2, 4, 1)

p1.add_atom('C')
p1.add_atom('N')
p1.add_atom('C')
p1.add_bond(1, 2, 1)
p1.add_bond(2, 3, 1)

p2.add_atom('C', _map=4)
p2.add_atom('Cl')
p2.add_bond(4, 5, 1)


# Synthesis of guanidines
#
q, p1, p2 = prepare()
q.add_atom('N')
q.add_atom('C')
q.add_atom('N', hybridization=1)
q.add_atom('N', hybridization=1)
q.add_bond(1, 2, 2)
q.add_bond(2, 3, 1)
q.add_bond(2, 4, 1)

p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('N')
p1.add_bond(1, 2, 3)
p1.add_bond(2, 3, 1)

p2.add_atom('N', _map=4)


# Grignard reaction with nitrile
#
q, p1, p2 = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('O')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 2)
q.add_bond(2, 4, 1)

p1.add_atom('C')
p1.add_atom('C')
p1.add_atom('N')
p1.add_bond(1, 2, 1)
p1.add_bond(2, 3, 3)

p2.add_atom('C', _map=4)
p2.add_atom('Br')
p2.add_bond(4, 5, 1)


# Alkylation of alpha-carbon atom of nitrile
#
q, p1, p2 = prepare()
q.add_atom('N')
q.add_atom('C')
q.add_atom('C', neighbors=(3, 4))
q.add_atom('C', hybridization=1)
q.add_bond(1, 2, 3)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)

p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('C')
p1.add_bond(1, 2, 3)
p1.add_bond(2, 3, 1)

p2.add_atom('C', _map=4)
p2.add_atom('Cl')
p2.add_bond(4, 5, 1)


# Gomberg-Bachmann reaction
#
q, p1, p2 = prepare()
q.add_atom('C', hybridization=4, heteroatoms=0)
q.add_atom('C', hybridization=4, heteroatoms=0)
q.add_bond(1, 2, 1)

p1.add_atom('C')
p1.add_atom('N', _map=3)
p1.add_bond(1, 3, 1)

p2.add_atom('C', _map=2)


# Cyclocondensation
#
q, p1, p2 = prepare()
q.add_atom('N', neighbors=2)
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('N')
q.add_atom('C')
q.add_atom('C')
q.add_atom('O', neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 2)
q.add_bond(5, 6, 1)
q.add_bond(6, 7, 1)
q.add_bond(7, 8, 2)
q.add_bond(1, 7, 1)

p1.add_atom('N')
p1.add_atom('C')
p1.add_atom('C')
p1.add_atom('C')
p1.add_atom('O', _map=9)
p1.add_bond(1, 2, 1)
p1.add_bond(2, 3, 1)
p1.add_bond(3, 4, 1)
p1.add_bond(4, 9, 2)

p2.add_atom('N', _map=5)
p2.add_atom('C')
p2.add_atom('C')
p2.add_atom('O')
p2.add_atom('O', _map=10)
p2.add_bond(5, 6, 1)
p2.add_bond(6, 7, 1)
p2.add_bond(7, 8, 2)
p2.add_bond(7, 10, 1)


# heterocyclization dicarboxylic acids
#
q, p1, p2 = prepare()
q.add_atom('C', rings_sizes=(5, 6))
q.add_atom('O')
q.add_atom(ListElement(['O', 'N']))
q.add_atom('C', rings_sizes=(5, 6))
q.add_atom('O')
q.add_bond(1, 2, 2)
q.add_bond(1, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 2)

p1.add_atom('C')
p1.add_atom('O')
p1.add_atom('O', _map=6)
p1.add_bond(1, 2, 2)
p1.add_bond(1, 6, 1)

p2.add_atom('C', _map=4)
p2.add_atom('O')
p2.add_atom('O', _map=7)
p2.add_bond(4, 5, 2)
p2.add_bond(4, 7, 1)


__all__ = ['rules']



from CGRtools import smiles
from ThetaSynthesis import RetroTree
from ThetaSynthesis.synthon import DoubleHeadedSynthon
from pickle import dump

target = smiles('O=[N+]([O-])c2oc(/C=N/N1C(=O)NC(=O)C1)cc2')
target.canonicalize()
target.clean2d()

tree = RetroTree(target, synthon_class=DoubleHeadedSynthon, size=10000, iterations=int(5e4))

with open('results.pkl', 'wb') as f:
    dump(list(tree), f)



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Tuple, Set, Iterable
from .abc import ScrollABC, IsTerminal
from .synthon.abc import SynthonABC


class Scroll(ScrollABC):
    __slots__ = ('_synthons', '_history', '_expand', '_closures', '_new_synthons')

    def __init__(self, synthons: Tuple[Tuple[SynthonABC, Set[SynthonABC]], ...],
                 new_synthons: Tuple[Tuple[SynthonABC, Set[SynthonABC]], ...], /):
        self._synthons = (*synthons, *(x for x in new_synthons if not x[0]))
        self._new_synthons = new_synthons

        if not self._synthons:
            self._expand: Iterable[SynthonABC] = ()
        else:
            self._expand = iter(self._synthons[0][0])

    def __call__(self, **kwargs):
        for synth, _ in self._new_synthons:
            synth(**kwargs)  # default scroll just transfer params into all new added synthons.

    @property
    def current_synthon(self):
        try:
            return self._synthons[0][0]
        except IndexError:
            raise IsTerminal

    @property
    def new_synthons(self):
        return tuple(x for x, _ in self._new_synthons)

    def __bool__(self):
        """
        Is terminal state. All synthons is building blocks
        """
        return not self._synthons

    def __len__(self):
        return len(self._synthons)

    def __float__(self):
        """
        Worse value from all synthons in the scroll
        """
        return min((float(x) for x, _ in self._synthons), default=1.)

    def __next__(self):
        """
        Expand Tree.
        """
        history = self._synthons[0][1]
        for prob, new in self._expand:
            if history.isdisjoint(new):
                return prob, type(self)(self._synthons[1:], tuple((x, {x, *history}) for x in new))
        raise StopIteration

    def __repr__(self):
        s = '\n'.join([repr(x) for x, _ in self._synthons])
        n = '\n'.join([repr(x) for x, _ in self._new_synthons])
        return f'queue:\n{s}\nnew:\n{n}'


__all__ = ['Scroll']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from typing import Dict, Set
from .scroll import ScrollABC


class RetroTreeABC(ABC):
    __slots__ = ('_succ', '_pred', '_nodes', '_free_node', '_visits', '_total_actions', '_probabilities')

    @abstractmethod
    def __init__(self, target: ScrollABC, /):
        self._succ: Dict[int, Set[int]] = {1: set()}
        self._pred: Dict[int, int] = {1: 0}
        self._nodes: Dict[int, ScrollABC] = {1: target}
        self._visits: Dict[int, int] = {1: 0}
        self._total_actions: Dict[int, float] = {1: 0.}
        self._probabilities: Dict[int, float] = {1: 0.}
        self._free_node: int = 2

    @abstractmethod
    def __next__(self) -> int:
        """
        Yield number of tree's node from building blocks.
        """

    def __iter__(self):
        return self

    def __len__(self):
        """
        Current size of tree
        """
        return self._free_node - 1

    def __repr__(self):
        return f'{type(self).__name__}({self._nodes[1]})'


__all__ = ['RetroTreeABC']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import MoleculeContainer, ReactionContainer
from itertools import count, islice
from math import sqrt
from tqdm import tqdm
from typing import Type, Optional, Tuple
from .abc import RetroTreeABC
from .scroll import Scroll
from .synthon.abc import SynthonABC


class RetroTree(RetroTreeABC):
    __slots__ = ('_depth', '_size', '_c_puct', '_expanded', '_iterations', '_limit', '_found', '_tqdm', '_node_depth')

    def __init__(self, target: MoleculeContainer, /, synthon_class: Type[SynthonABC],
                 c_puct: float = 4., depth: int = 10, size: int = 1e4, iterations: int = 1e5):
        """
        :param target: target molecule
        :param c_puct: breadth/depth criterion
        :param depth: max length of path to building blocks
        :param size: max size of tree
        :param iterations: limit of iterations
        """
        self._depth = depth
        self._size = int(size)
        self._c_puct = c_puct
        self._expanded = 1
        self._limit = iterations = int(iterations)
        self._iterations = 0
        self._found = 0
        self._node_depth = {1: 0}
        self._tqdm = tqdm(total=iterations)

        synthon = synthon_class(target)
        scroll = Scroll((), ((synthon, {synthon}),))
        scroll(finish=self._depth)
        super().__init__(scroll)

    def synthesis_path(self, node: int) -> Tuple[ReactionContainer, ...]:
        """
        Return a synthesis path from a picked node's molecule to target molecule.

        :param node: building block node
        """
        nodes = self.chain_to_node(node)

        tmp = [ReactionContainer([x.molecule for x in after.new_synthons], [before.current_synthon.molecule])
               for before, after in zip(nodes, nodes[1:])]
        for r in tmp:
            r.clean2d()
        return tuple(reversed(tmp))

    def chain_to_node(self, node: int):
        nodes = []
        while node:
            nodes.append(node)
            node = self._pred[node]
        return [self._nodes[node] for node in reversed(nodes)]

    def path_graph(self, node: int) -> str:
        """
        Visualize reaction path.

        :param node: building block node
        :return: SVG string
        """
        nodes = []
        while node:
            nodes.append(node)
            node = self._pred[node]
        nodes = [self._nodes[node] for node in reversed(nodes)]
        # first column is target
        # second column are first new synthons
        columns = [[nodes[0].current_synthon.molecule],
                   [x.molecule for x in nodes[1].new_synthons]]
        pred = {x: 0 for x in range(1, len(columns[1]) + 1)}
        cx = [n for n, x in enumerate(nodes[1].new_synthons, 1) if not x]
        size = len(cx)
        nodes = iter(nodes[2:])
        cy = count(len(columns[1]) + 1)
        while size:
            layer = []
            for s in islice(nodes, size):
                n = cx.pop(0)
                for x in s.new_synthons:
                    layer.append(x)
                    m = next(cy)
                    if not x:
                        cx.append(m)
                    pred[m] = n
            size = len(cx)
            columns.append([x.molecule for x in layer])

        # now we have columns for visualizing
        # lets start recalculate XY
        x_shift = 0.
        c_max_x = 0.
        c_max_y = 0.
        render = []
        cx = count()
        cy = count()
        arrow_points = {}
        for ms in columns:
            heights = []
            for m in ms:
                m.clean2d()
                # X-shift for molecule
                min_x = min(x for x, y in m._plane.values()) - x_shift
                min_y = min(y for x, y in m._plane.values())
                m._plane = {n: (x - min_x, y - min_y) for n, (x, y) in m._plane.items()}
                max_x = max(x for x, y in m._plane.values())
                if max_x > c_max_x:
                    c_max_x = max_x
                arrow_points[next(cx)] = [x_shift, max_x]
                heights.append(max(y for x, y in m._plane.values()))
            x_shift = c_max_x + 5.  # between columns gap
            # calculate Y-shift
            y_shift = (sum(heights) + 3. * (len(heights) - 1))
            if y_shift > c_max_y:
                c_max_y = y_shift
            y_shift /= 2.
            for m, h in zip(ms, heights):
                m._plane = {n: (x, y - y_shift) for n, (x, y) in m._plane.items()}
                arrow_points[next(cy)].append(y_shift - h / 2.)
                y_shift -= h + 3.
                render.append(m.depict(embedding=True)[:3])

        config = MoleculeContainer._render_config
        font_size = config['font_size']
        font125 = 1.25 * font_size
        width = c_max_x + 3.0 * font_size
        height = c_max_y + 2.5 * font_size
        box_y = height / 2.
        svg = [f'<svg width="{width:.2f}cm" height="{height:.2f}cm" '
               f'viewBox="{-font125:.2f} {-box_y:.2f} {width:.2f} '
               f'{height:.2f}" xmlns="http://www.w3.org/2000/svg" version="1.1">',
               '  <defs>\n    <marker id="arrow" markerWidth="10" markerHeight="10" '
               'refX="0" refY="3" orient="auto">\n      <path d="M0,0 L0,6 L9,3"/>\n    </marker>\n  </defs>']

        for s, p in pred.items():
            s_min_x, s_max, s_y = arrow_points[s]
            p_min_x, p_max, p_y = arrow_points[p]
            svg.append(f'  <line x1="{p_max + 1.:.2f}" y1="{p_y:.2f}" x2="{s_min_x - 1.:.2f}" y2="{s_y:.2f}" '
                       'fill="none" stroke="black" stroke-width=".04" marker-end="url(#arrow)"/>')
        for atoms, bonds, masks in render:
            svg.extend(MoleculeContainer._graph_svg(atoms, bonds, masks, -font125, -box_y, width, height))
        svg.append('</svg>')
        return '\n'.join(svg)

    def find_target(self, molecule: 'MoleculeContainer'):
        nodes = []
        for idx, node in self._nodes.items():
            if any(synth and molecule == synth._molecule for synth in node.new_synthons):
                nodes.append(idx)
            elif not node and node.current_synthon._molecule == molecule:
                nodes.append(idx)
        return nodes

    def report(self) -> str:
        return f'Tree for: {self._nodes[1]}\n' \
               f'Size: {len(self)}\nNumber of unvisited nodes: {self._free_node - self._expanded}\n' \
               f'Found paths: {self._found}'

    def visualize(self, draw_format: str = 'png', prog: Optional[str] = None,
                  only_visited: bool = False, verbose: int = 2):
        import pygraphviz as pgv
        from warnings import warn

        lst = ['id', 'visits', 'smiles in queue', 'value']
        if verbose == 3 and not only_visited:
            warn('Verbose = 3 option can be used only for visited nodes', UserWarning)
            only_visited = True
        elif verbose == 2:
            lst = lst[:3]
        elif verbose == 1:
            lst = lst[:2]
        elif verbose == 0:
            lst = lst[:1]

        if only_visited:
            nodes = {k: v for k, v in self._nodes.items() if self._visits[k]}
        else:
            nodes = {k: v for k, v in self._nodes.items()}

        lambdas = [lambda x: x, lambda x: self._visits[x], lambda x: repr(x), lambda x: float(self._nodes[x])]

        nodes_with_attrs = {
            k: '\n'.join(f'{x}: {z(y)}' for x, y, z in zip(lst, [k, k, v, k], lambdas))
            for k, v
            in nodes.items()
        }
        pred = self._pred
        g = pgv.AGraph(directed=True)
        g.node_attr['shape'] = 'box'
        g.add_edges_from([(v, k) for k, v in pred.items() if k != 1 and k in nodes])

        for k, node in zip(nodes, g.nodes()):
            node.attr['label'] = nodes_with_attrs[k]

        g.layout(prog='dot')
        return g.draw(format=draw_format, prog=prog)

    def _add(self, node: int, scroll: Scroll, prob: float):
        """
        Add new node to tree.
        """
        new_node = self._free_node
        self._nodes[new_node] = scroll
        self._pred[new_node] = node
        self._succ[node].add(new_node)
        self._succ[new_node] = set()
        self._visits[new_node] = 0
        self._probabilities[new_node] = prob
        self._total_actions[new_node] = 0.
        self._node_depth[new_node] = self._node_depth[node] + 1
        self._free_node += 1

    def _update_visits(self, node: int):
        """
        Increment visits count in path from given node to root.
        """
        while node:
            self._visits[node] += 1
            node = self._pred[node]

    def _update_actions(self, node: int):
        """
        Update total action of each node in path to root by value of given node.
        """
        value = float(self._nodes[node])
        while node:
            self._total_actions[node] += value
            node = self._pred[node]

    def _expand(self, node: int):
        """
        Expand new node.
        """
        finish = self._depth - self._node_depth[node]
        for prob, scroll in self._nodes[node]:
            scroll(finish=finish)  # init new scroll
            self._add(node, scroll, prob)

    def _select(self, node: int) -> int:
        """
        Select preferred successor node based on views count and synthesizability.
        """
        return max(self._succ[node], key=self._puct)

    def _puct(self, node):
        """
        Polynomial upper confidence trees criterion for choosing node from tree
        """
        prob = self._probabilities[node]
        visit = self._visits[node]

        # C_PUCT is a constant determining a level of exploration; can be from 1 to 6; 4 is more balanced value
        u = self._c_puct * prob * sqrt(sum(self._visits[x] for x in self._succ[self._pred[node]]) + 1)

        return (self._total_actions[node] + u) / (visit + 1)

    def __next__(self):
        while self._expanded < self._free_node:
            self._iterations += 1
            if self._iterations > self._limit:
                self._tqdm.close()
                raise StopIteration('Iterations limit exceeded. \n' + self.report())
            self._tqdm.update()
            depth = 0
            node = 1
            while True:
                if self._visits[node]:  # already expanded
                    if not self._succ[node]:  # dead terminal non-building block node.
                        self._update_visits(node)
                        break
                    node = self._select(node)
                    depth += 1
                else:
                    self._expanded += 1  # increment visited nodes count.
                    if self._nodes[node]:  # found path!
                        self._update_visits(node)  # this prevents expanding of bb node
                        # I dunno: self._update_actions(node)
                        self._found += 1
                        return node
                    elif depth < self._depth and self._free_node < self._size:  # expand if depth limit not reached
                        self._expand(node)
                        self._update_visits(node)  # mark node as visited
                        self._update_actions(node)
                        break
                    else:
                        self._update_visits(node)
                        self._update_actions(node)
                        break
        self._tqdm.close()
        raise StopIteration('Max tree size exceeded or all possible paths found' + self.report())

    def __repr__(self):
        return self.report()


__all__ = ['RetroTree']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import Reactor, SMILESRead
from io import TextIOWrapper
from itertools import takewhile
from torch import hstack, Tensor
from pickle import load
from pkg_resources import resource_stream
from StructureFingerprint import LinearFingerprint
from torch import from_numpy, sort
from .model import FilterNet, SorterNet, DoubleHeadedNet
from ..rollout import RolloutSynthon


class PolicySynthon(RolloutSynthon):
    __slots__ = ('_bit_string', )
    __filter__ = None
    __sorter__ = None
    __fragmentor__ = None

    __filter_path__ = None
    __sorter_path__ = None

    def __new__(cls, molecule, *args, **kwargs):
        if cls.__bb__ is None:
            with resource_stream(__name__, 'data/rules.pickle') as f:
                rules = load(f)
            cls.__bb__ = frozenset(str(m) for m in SMILESRead(TextIOWrapper(resource_stream(__name__, 'data/bb.smi'))))
            cls.__reactors__ = tuple(Reactor(x, delete_atoms=True) for x in rules)
            cls.__fragmentor__ = LinearFingerprint(length=4096, min_radius=2, max_radius=4, number_bit_pairs=4)
        if cls.__filter__ is None:
            cls.__filter__ = FilterNet().load_from_checkpoint(cls.__filter_path__)
            cls.__filter__.eval()
            cls.__sorter__ = SorterNet().load_from_checkpoint(cls.__sorter_path__)
            cls.__sorter__.eval()
        return super().__new__(cls, molecule, *args, **kwargs)

    def __init__(self, molecule, /):
        super().__init__(molecule)
        self._bit_string = from_numpy(self.__fragmentor__.transform([self._molecule])[0]).float()

    def _sorted(self):
        reactors = self.__reactors__
        vec = self._bit_string
        res = (self.__sorter__.forward(vec) + .01) * self.__filter__._predict(vec)
        sorted_, values = sort(res.sqrt(), descending=True)
        yield from ((x.item(), reactors[y.item()])
                    for x, y in zip(sorted_.squeeze(), values.squeeze())
                    if x > 0.)


class DoubleHeadedSynthon(PolicySynthon):
    __slots__ = ('_probs', )
    __net__ = None

    def __new__(cls, molecule, *args, **kwargs):
        if cls.__net__ is None:
            cls.__net__ = DoubleHeadedNet()
            cls.__net__.eval()
        return super().__new__(cls, molecule, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        super().__call__(*args, **kwargs)
        self._probs, float = self.__net__.forward(hstack([self._bit_string, Tensor([self._depth])]))
        self._float = float.item()

    def __float__(self):
        if self._float is not None:
            return self._float
        elif self:
            self._float = 1.
            return self._float

    def _sorted(self):
        reactors = self.__reactors__
        sorted_, values = sort(self._probs, descending=True)
        yield from ((x.item(), reactors[y.item()])
                    for x, y in takewhile(lambda x: x[0] > .1, zip(sorted_.squeeze(), values.squeeze())))


__all__ = ['PolicySynthon', 'DoubleHeadedSynthon']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import smiles
from .abc import SynthonABC


class DummySynthon(SynthonABC):
    """
    Test synthon for Acetaminophen.
    """
    __slots__ = ()

    def __iter__(self):
        for prob, molecules in data[self._molecule][0]:
            yield prob, tuple(type(self)(mol) for mol in molecules)

    def __bool__(self):
        return self._molecule in building_blocks

    def __float__(self):
        return data[self._molecule][1]


building_blocks = {smiles('Oc1ccccc1')}
pre_data = {
    'CC(=O)Nc1ccc(O)cc1': (((.25, ('Oc1ccc(N)cc1',),),
                            (.15, ('C(Nc1ccc(OC)cc1)(C)=O',)),
                            (.2, ('ON=C(C)c1ccc(O)cc1',)),
                            (.25, ('Oc1ccc(O)cc1',)),
                            (.15, ('C(Nc1ccc(OC2OCCCC2)cc1)(C)=O',))), 1.),
    'Oc1ccc(N)cc1': (((.3, ('O=N(=O)c1ccc(O)cc1',)),
                      (.4, ('c1(N)ccc(OC)cc1',)),
                      (.3, ('c1(O)ccc(F)cc1',))), 1.),
    'O=N(=O)c1ccc(O)cc1': (((.35, ('Oc1ccccc1',)),
                            (.35, ('N(c1ccc(N)cc1)(=O)=O',)),
                            (.3, ('c1(N(=O)=O)cc(c(cc1)O)Br',))), 1.),
    'Oc1ccccc1': ((), 1.),
    'C(Nc1ccc(OC)cc1)(C)=O': ((), -1.),
    'Oc1ccc(O)cc1': ((), 0.),
    'C(Nc1ccc(OC2OCCCC2)cc1)(C)=O': ((), -1.),
    'c1(N)ccc(OC)cc1': ((), -1.),
    'c1(O)ccc(F)cc1': (((1., ('c1(OCOCCOC)ccc(F)cc1',)),), .1),
    'c1(OCOCCOC)ccc(F)cc1': ((), -1.),
    'c1(N(=O)=O)cc(c(cc1)O)Br': ((), -1.),
    'N(c1ccc(N)cc1)(=O)=O': (((1., ('C(c1c([N+](=O)[O-])ccc(c1)N)(=O)O',)),), -.1),
    'C(c1c([N+](=O)[O-])ccc(c1)N)(=O)O': ((), -1.),
    'ON=C(C)c1ccc(O)cc1': (((.5, ('O=C(C)c1ccc(O)cc1',)),
                            (.2, ('[Si](C(C)(C)C)(Oc1ccc(C(=NO)C)cc1)(C)C',)),
                            (.3, ('c1(c(ccc(C(=NO)C)c1)O)N',))), 1.),
    'c1(c(ccc(C(=NO)C)c1)O)N': (((1., ('c1(c(cc(C(=NO)C)cc1)N)OCOC',)),), -.2),
    'c1(c(cc(C(=NO)C)cc1)N)OCOC': ((), -1.),
    'O=C(C)c1ccc(O)cc1': (((.85, ('Oc1ccccc1',)),
                           (.15, ('[Si](C(C)(C)C)(Oc1ccc(C(=O)C)cc1)(C)C',))), 1.),
    '[Si](C(C)(C)C)(Oc1ccc(C(=O)C)cc1)(C)C': ((), -1.),
    '[Si](C(C)(C)C)(Oc1ccc(C(=NO)C)cc1)(C)C': ((), -1.),
}


def convert(dct):
    out = {}
    for k, (tuples, value) in dct.items():
        k = smiles(k)
        k.canonicalize()
        new_mols = []
        for prob, mols in tuples:
            synth = []
            for mol in mols:
                new_mol = smiles(mol)
                new_mol.canonicalize()
                synth.append(new_mol)
            new_mols.append((prob, tuple(synth)))
        out[k] = (tuple(new_mols), value)
    return out


data = convert(pre_data)


__all__ = ['DummySynthon']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CGRtools import QueryContainer, ReactionContainer
from CGRtools.periodictable import ListElement


rules = []


def prepare():
    q_ = QueryContainer()
    p_ = QueryContainer()
    rules.append(ReactionContainer((q_,), (p_,)))
    return q_, p_


# aryl nitro reduction
# [C;Za;W1]-[N;D1]>>[O-]-[N+](-[C])=[O]
q, p = prepare()
q.add_atom('N', neighbors=1)
q.add_atom('C', hybridization=4, heteroatoms=1)
q.add_bond(1, 2, 1)

p.add_atom('N', charge=1)
p.add_atom('C')
p.add_atom('O', charge=-1)
p.add_atom('O')
p.add_bond(1, 2, 1)
p.add_bond(1, 3, 1)
p.add_bond(1, 4, 2)


# aryl nitration
# [O-]-[N+](=[O])-[C;Za;W12]>>[C]
q, p = prepare()
q.add_atom('N', charge=1)
q.add_atom('C', hybridization=4, heteroatoms=(1, 2))
q.add_atom('O', charge=-1)
q.add_atom('O')
q.add_bond(1, 2, 1)
q.add_bond(1, 3, 1)
q.add_bond(1, 4, 2)

p.add_atom('C', _map=2)


# Beckmann rearrangement (oxime -> amide)
# [C]-[N;D2]-[C]=[O]>>[O]-[N]=[C]-[C]
q, p = prepare()
q.add_atom('C')
q.add_atom('N', neighbors=2)
q.add_atom('O')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(1, 3, 2)
q.add_bond(2, 4, 1)

p.add_atom('C')
p.add_atom('N')
p.add_atom('O')
p.add_atom('C')
p.add_bond(1, 2, 2)
p.add_bond(2, 3, 1)
p.add_bond(1, 4, 1)


# aldehydes or ketones into oxime/imine reaction
# [C;Zd;W1]=[N]>>[C]=[O]
q, p = prepare()
q.add_atom('C', hybridization=2, heteroatoms=1)
q.add_atom('N')
q.add_bond(1, 2, 2)

p.add_atom('C')
p.add_atom('O', _map=3)
p.add_bond(1, 3, 2)


# addition of halogen atom into phenol ring (orto)
# [C](-[Cl,F,Br,I;D1]):[C]-[O,N;Zs]>>[C](-[A]):[C]
q, p = prepare()
q.add_atom(ListElement(['O', 'N']), hybridization=1)
q.add_atom('C')
q.add_atom('C')
q.add_atom(ListElement(['Cl', 'F', 'Br', 'I']), neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 4)
q.add_bond(3, 4, 1)

p.add_atom('A')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 4)


# addition of halogen atom into phenol ring (para)
# [C](:[C]:[C]:[C]-[O,N;Zs])-[Cl,F,Br,I;D1]>>[A]-[C]:[C]:[C]:[C]
q, p = prepare()
q.add_atom(ListElement(['O', 'N']), hybridization=1)
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom(ListElement(['Cl', 'F', 'Br', 'I']), neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 4)
q.add_bond(3, 4, 4)
q.add_bond(4, 5, 4)
q.add_bond(5, 6, 1)

p.add_atom('A')
p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 4)
p.add_bond(3, 4, 4)
p.add_bond(4, 5, 4)


# hard reduction of Ar-ketones
# [C;Za]-[C;D2;Zs;W0]>>[C]-[C]=[O]
q, p = prepare()
q.add_atom('C', hybridization=4)
q.add_atom('C', hybridization=1, neighbors=2, heteroatoms=0)
q.add_bond(1, 2, 1)

p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 2)


# reduction of alpha-hydroxy pyridine
# [C;W1]:[N;H0;r6]>>[C](:[N])-[O]
q, p = prepare()
q.add_atom('C', heteroatoms=1)
q.add_atom('N', rings_sizes=6, hydrogens=0)
q.add_bond(1, 2, 4)

p.add_atom('C')
p.add_atom('N')
p.add_atom('O')
p.add_bond(1, 2, 4)
p.add_bond(1, 3, 1)


# Reduction of alkene
# [C]-[C;D23;Zs;W0]-[C;D123;Zs;W0]>>[C](-[C])=[C]
q, p = prepare()
q.add_atom('C')
q.add_atom('C', heteroatoms=0, neighbors=(2, 3), hybridization=1)
q.add_atom('C', heteroatoms=0, neighbors=(1, 2, 3), hybridization=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)

p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 2)


# Kolbe-Schmitt reaction
# [C](:[C]-[O;D1])-[C](=[O])-[O;D1]>>[C](-[O]):[C]
q, p = prepare()
q.add_atom('O', neighbors=1)
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('O', neighbors=1)
q.add_atom('O')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 4)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 1)
q.add_bond(4, 6, 2)

p.add_atom('O')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 4)


# reduction of carboxylic acid
# [O;D1]-[C;D2]-[C]>>[C]-[C](-[O])=[O]
q, p = prepare()
q.add_atom('C')
q.add_atom('C', neighbors=2)
q.add_atom('O', neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)

p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_atom('O')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 1)
p.add_bond(2, 4, 2)


# halogenation of alcohols
# [C;Zs]-[Cl,Br;D1]>>[C]-[O]
q, p = prepare()
q.add_atom('C', hybridization=1, heteroatoms=1)
q.add_atom(ListElement(['Cl', 'Br']), neighbors=1)
q.add_bond(1, 2, 1)

p.add_atom('C')
p.add_atom('O', _map=3)
p.add_bond(1, 3, 1)


# Kolbe nitrilation
# [N]#[C]-[C;Zs;W0]>>[Br]-[C]
q, p = prepare()
q.add_atom('C', heteroatoms=0, hybridization=1)
q.add_atom('C')
q.add_atom('N')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 3)

p.add_atom('C')
p.add_atom('Br', _map=4)
p.add_bond(1, 4, 1)


# Nitrile hydrolysis
# [O;D1]-[C]=[O]>>[N]#[C]
q, p = prepare()
q.add_atom('C')
q.add_atom('O', neighbors=1)
q.add_atom('O')
q.add_bond(1, 2, 1)
q.add_bond(1, 3, 2)

p.add_atom('C')
p.add_atom('N', _map=4)
p.add_bond(1, 4, 3)


# sulfamidation
# [c]-[S](=[O])(=[O])-[N]>>[c]
q, p = prepare()
q.add_atom('C', hybridization=4)
q.add_atom('S')
q.add_atom('O')
q.add_atom('O')
q.add_atom('N', neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 2)
q.add_bond(2, 4, 2)
q.add_bond(2, 5, 1)

p.add_atom('C')


# Ring expansion rearrangement
#
q, p = prepare()
q.add_atom('C')
q.add_atom('N')
q.add_atom('C', rings_sizes=6)
q.add_atom('C')
q.add_atom('O')
q.add_atom('C')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 2)
q.add_bond(3, 6, 1)
q.add_bond(4, 7, 1)

p.add_atom('C')
p.add_atom('N')
p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 1)
p.add_bond(2, 3, 2)
p.add_bond(3, 4, 1)
p.add_bond(4, 5, 1)
p.add_bond(4, 6, 1)
p.add_bond(4, 7, 1)


# hydrolysis of bromide alkyl
#
q, p = prepare()
q.add_atom('C', hybridization=1)
q.add_atom('O', neighbors=1)
q.add_bond(1, 2, 1)

p.add_atom('C')
p.add_atom('Br')
p.add_bond(1, 2, 1)


# Condensation of ketones/aldehydes and amines into imines
#
q, p = prepare()
q.add_atom('N', neighbors=(1, 2))
q.add_atom('C', neighbors=(2, 3), heteroatoms=1)
q.add_bond(1, 2, 2)

p.add_atom('C', _map=2)
p.add_atom('O')
p.add_bond(2, 3, 2)


# Halogenation of alkanes
#
q, p = prepare()
q.add_atom('C', hybridization=1)
q.add_atom(ListElement(['F', 'Cl', 'Br']))
q.add_bond(1, 2, 1)

p.add_atom('C')


# heterocyclization
#
q, p = prepare()
q.add_atom('N', heteroatoms=0, hybridization=1, neighbors=(2, 3))
q.add_atom('C', heteroatoms=2)
q.add_atom('N', heteroatoms=0, neighbors=2)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 2)

p.add_atom('N')
p.add_atom('C')
p.add_atom('N')
p.add_atom('O')
p.add_bond(1, 2, 1)
p.add_bond(2, 4, 2)


# Reduction of nitrile
#
q, p = prepare()
q.add_atom('N', neighbors=1)
q.add_atom('C')
q.add_atom('C', hybridization=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)

p.add_atom('N')
p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 3)
p.add_bond(2, 3, 1)


# SPECIAL CASE
# Reduction of nitrile into methylamine
#
q, p = prepare()
q.add_atom('C', neighbors=1)
q.add_atom('N', neighbors=2)
q.add_atom('C')
q.add_atom('C', hybridization=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)

p.add_atom('N', _map=2)
p.add_atom('C')
p.add_atom('C')
p.add_bond(2, 3, 3)
p.add_bond(3, 4, 1)


# methylation of amides
#
q, p = prepare()
q.add_atom('O')
q.add_atom('C')
q.add_atom('N')
q.add_atom('C', neighbors=1)
q.add_bond(1, 2, 2)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)

p.add_atom('O')
p.add_atom('C')
p.add_atom('N')
p.add_bond(1, 2, 2)
p.add_bond(2, 3, 1)


# hydrocyanation of alkenes
#
q, p = prepare()
q.add_atom('C', hybridization=1)
q.add_atom('C')
q.add_atom('C')
q.add_atom('N')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 3)

p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 2)


# decarbocylation (alpha atom of nitrile)
#
q, p = prepare()
q.add_atom('N')
q.add_atom('C')
q.add_atom('C', neighbors=2)
q.add_bond(1, 2, 3)
q.add_bond(2, 3, 1)

p.add_atom('N')
p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_atom('O')
p.add_bond(1, 2, 3)
p.add_bond(2, 3, 1)
p.add_bond(3, 4, 1)
p.add_bond(4, 5, 2)
p.add_bond(4, 6, 1)


# Bichler-Napieralski reaction
#
q, p = prepare()
q.add_atom('C', rings_sizes=(6,))
q.add_atom('C', rings_sizes=(6,))
q.add_atom('N', rings_sizes=(6,), neighbors=2)
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom('O')
q.add_atom('O')
q.add_atom('C')
q.add_atom('O', neighbors=1)
q.add_bond(1, 2, 4)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 2)
q.add_bond(5, 6, 1)
q.add_bond(6, 7, 2)
q.add_bond(6, 8, 1)
q.add_bond(5, 9, 4)
q.add_bond(9, 10, 1)
q.add_bond(1, 9, 1)

p.add_atom('C')
p.add_atom('C')
p.add_atom('N')
p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_atom('O')
p.add_atom('C')
p.add_atom('O')
p.add_atom('O')
p.add_bond(1, 2, 4)
p.add_bond(2, 3, 1)
p.add_bond(3, 4, 1)
p.add_bond(4, 5, 2)
p.add_bond(5, 6, 1)
p.add_bond(6, 7, 2)
p.add_bond(6, 8, 1)
p.add_bond(5, 9, 1)
p.add_bond(9, 10, 2)
p.add_bond(9, 11, 1)


# heterocyclization in Prins reaction
#
q, p = prepare()
q.add_atom('C')
q.add_atom('O')
q.add_atom('C')
q.add_atom(ListElement(['N', 'O']), neighbors=2)
q.add_atom('C')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 1)
q.add_bond(5, 6, 1)
q.add_bond(1, 6, 1)

p.add_atom('C')
p.add_atom('C', _map=5)
p.add_bond(1, 5, 2)


# recyclization of tetrahydropyran through an opening the ring and dehydration
#
q, p = prepare()
q.add_atom('C')
q.add_atom('C')
q.add_atom('C')
q.add_atom(ListElement(['N', 'O']))
q.add_atom('C')
q.add_atom('C')
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)
q.add_bond(3, 4, 1)
q.add_bond(4, 5, 1)
q.add_bond(5, 6, 1)
q.add_bond(1, 6, 2)

p.add_atom('C')
p.add_atom('C')
p.add_atom('C')
p.add_atom('A')
p.add_atom('C')
p.add_atom('C')
p.add_atom('O')
p.add_bond(1, 2, 1)
p.add_bond(1, 7, 1)
p.add_bond(3, 7, 1)
p.add_bond(3, 4, 1)
p.add_bond(4, 5, 1)
p.add_bond(5, 6, 1)
p.add_bond(1, 6, 1)


# alkenes + h2o/hHal
#
q, p = prepare()
q.add_atom('C', hybridization=1)
q.add_atom('C', hybridization=1)
q.add_atom(ListElement(['O', 'F', 'Cl', 'Br', 'I']), neighbors=1)
q.add_bond(1, 2, 1)
q.add_bond(2, 3, 1)

p.add_atom('C')
p.add_atom('C')
p.add_bond(1, 2, 2)


# methylation of dimethylamines
#
q, p = prepare()
q.add_atom('C', neighbors=1)
q.add_atom('N', neighbors=3)
q.add_bond(1, 2, 1)

p.add_atom('N', _map=2)


__all__ = ['rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .oneheaded import *
from .double_headed import *


__all__ = ['FilterNet', 'SorterNet', 'DoubleHeadedNet']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from typing import Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..synthon.abc import SynthonABC


class IsTerminal(Exception):
    """Scroll is terminal"""


class ScrollABC(ABC):
    """
    Node of MCTS Tree
    """
    __slots__ = ()

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self) -> Tuple[float, 'ScrollABC']:
        """
        Yield pairs of reaction value and Scroll.
        """

    @abstractmethod
    def __bool__(self):
        ...

    @abstractmethod
    def __float__(self):
        ...

    @property
    @abstractmethod
    def current_synthon(self) -> 'SynthonABC':
        """
        Return a synthon from the top of the queue of node's synthons.
        raise IsTerminal exception then scroll do not contains non-building blocks.
        """

    @property
    @abstractmethod
    def new_synthons(self) -> Tuple['SynthonABC', ...]:
        """
        Return a collection of new just added synthons.
        """

    @abstractmethod
    def __call__(self, **kwargs):
        """
        Apply additional params from tree to scroll.

        Unified way for tree customizations.
        """


__all__ = ['ScrollABC', 'IsTerminal']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from pathlib import Path
from setuptools import setup


setup(
    name='ThetaSynthesis',
    version='0.1.0',
    packages=['ThetaSynthesis', 'ThetaSynthesis.abc', 'ThetaSynthesis.synthon', 'ThetaSynthesis.synthon.abc',
              'ThetaSynthesis.synthon.rollout', 'ThetaSynthesis.synthon.rollout.rules',
              'ThetaSynthesis.synthon.policy', 'ThetaSynthesis.synthon.policy.model'],
    python_requires='>=3.8.1',
    install_requires=['CGRtools>=4.1,<4.2', 'tqdm', 'StructureFingerprint',
                      'torch', 'pytorch-lightning', 'torchmetrics'],
    package_data={'ThetaSynthesis.synthon.rollout': ['data/*'],
                  'ThetaSynthesis.synthon.policy': ['data/*'],
                  'ThetaSynthesis.synthon.policy.model': ['data/*']},
    zip_safe=True,
    license='LGPLv3',
    url='https://github.com/dcloudf/ThetaSynthesis',
    author='Alexander Sizov',
    author_email='',
    long_description=(Path(__file__).parent / 'README.rst').read_text(),
    classifiers=['Environment :: Plugins',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3 :: Only',
                 'Programming Language :: Python :: 3.8',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Software Development',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Software Development :: Libraries :: Python Modules']
)



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Alexander Sizov <murkyrussian@gmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import chain, repeat
from typing import Any, Union, List, Optional, TYPE_CHECKING, Iterable

import torch

from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from ....tree import RetroTree
from .. import DoubleHeadedSynthon

if TYPE_CHECKING:
    from CGRtools import MoleculeContainer


class RetroTreeDataModule(LightningDataModule):
    def prepare_data(self, *args, **kwargs):
        pass

    def setup(self, stage: Optional[str] = None):
        pass

    def train_dataloader(self, *args, **kwargs) -> DataLoader:
        pass

    def val_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        pass

    def test_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        pass

    def transfer_batch_to_device(self, batch: Any, device: Optional[torch.device] = None) -> Any:
        pass

    @staticmethod
    def __collect_from_tree(x: Iterable['MoleculeContainer']):
        for mol in x:
            tree = RetroTree(mol, synthon_class=DoubleHeadedSynthon)
            win_terminals = list(tree)
            winner_nodes = list(zip(chain.from_iterable([tree.chain_to_node(node) for node in win_terminals]), repeat(1.)))
            loser_nodes = zip(
                sorted(tree._nodes,
                       key=lambda k: tree._visits[k],
                       reverse=True)[:min(len(winner_nodes), 10)],
                repeat(-1.))
            for example in chain(winner_nodes, loser_nodes):
                node, value = example
                yield [tree._nodes[node].current_synthon._bit_string, ..., ...]


__all__ = ['RetroTreeDataModule']



# -*- coding: utf-8 -*-
#
#  Copyright 2020-2021 Alexander Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of ThetaSynthesis.
#
#  ThetaSynthesis is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from pkg_resources import resource_stream
from torch import hstack, Tensor
from torch.nn import Linear, Sequential, Softmax
from torch.nn.functional import kl_div, mse_loss
from torch.optim import Adam
from . import SorterNet


class DoubleHeadedNet(SorterNet):
    def __init__(self):
        super().__init__()

        self.policy_net = SorterNet.load_from_checkpoint(resource_stream(__name__, 'data/sorter.ckpt'))

        self.body = self.policy_net.body

        self.policy_head = Softmax(dim=0)

        self.value_head = Sequential(
            Linear(2273, 1),
        )

    def forward(self, x):
        if isinstance(x, Tensor) and x.shape[0] == 4096:
            return self.policy_head(self.body(x))
        elif isinstance(x, Tensor) and x.shape[0] == 4097:
            x_policy, x_value = x[:-1], x[-1]
        elif isinstance(x, tuple):
            x_policy, x_value = x
        else:
            raise TypeError

        policy = self.policy_net(x_policy)
        stack = hstack((policy, x_value))

        value = self.value_head(stack)
        return policy, value

    def _predict(self, x):
        return self.forward(x)

    def training_step(self, batch, batch_idx):
        loss_policy, loss_value = self._losses(batch, batch_idx)
        opt = self.optimizers()

        self.manual_backward(loss_policy, opt, retain_graph=True)
        self.manual_backward(loss_value, opt)

        self.log('loss_policy', loss_policy)
        self.log('loss_value', loss_value)

        loss = loss_policy + loss_value
        self.log('loss', loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss_policy, loss_value = self._losses(batch, batch_idx)

        self.log('val_loss_policy', loss_policy)
        self.log('val_loss_value', loss_value)

        loss = loss_policy + loss_value
        self.log('val_loss', loss)
        return loss

    def test_step(self, batch, batch_idx):
        loss_policy, loss_value = self._losses(batch, batch_idx)

        self.log('test_loss_policy', loss_policy)
        self.log('test_loss_value', loss_value)

        loss = loss_policy + loss_value
        self.log('test_loss', loss)
        return loss

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=3e-4)
        return optimizer

    def _losses(self, batch, batch_idx):
        x, y = batch

        finger, depth = x[:, :-1], x[:, -1].reshape(-1, 1)
        y_policy, y_value = y[:, :-1], y[:, -1].reshape(-1, 1)

        pred_policy = self.policy_net(finger)
        stack = hstack((self.body(finger), depth))

        pred_value = self.value_head(stack)

        loss_policy = kl_div(pred_policy, y_policy, reduction='batchmean')
        loss_value = mse_loss(pred_value, y_value)

        return loss_policy, loss_value

    def _positive_exps_learn(self):
        ...

    def _negative_exps_learn(self):
        ...


__all__ = ['DoubleHeadedNet']
